@@ -145,12 +145,13 @@ sub handle_put {
     # return a 400 (bad request) if we got no content length or if it's
     # bigger than any specified max put size
     return $self->send_response(400, "Content-length of $clen is invalid.")
-        if !$clen ||
+        if ! defined($clen) ||
+        $clen < 0 ||
         ($self->{service}->{max_put_size} &&
          $clen > $self->{service}->{max_put_size});
 
-    # if we have some data already from a header over-read, note it
-    if (defined $self->{read_ahead} && $self->{read_ahead} > 0) {
+    # if we are supposed to read data and have some data already from a header over-read, note it
+    if ($clen && defined $self->{read_ahead} && $self->{read_ahead} > 0) {
         $self->{content_length_remain} -= $self->{read_ahead};
     }
 
@@ -227,7 +228,6 @@ sub handle_put_chunked {
     my ($path, $filename) = ($1 || '', $2);
 
     my $disk_path = $self->{service}->{docroot} . '/' . $path;
-    $self->start_put_open($disk_path, $filename);
 
     $self->{chunked_upload_state} = Perlbal::ChunkedUploadState->new(%{{
         on_new_chunk => sub {
@@ -247,6 +247,9 @@ sub handle_put_chunked {
                 return;
             }
 
+            # Reading too far ahead of our AIO subsystem will cause us to buffer it in memory.
+            $self->watch_read(0) if $self->{read_ahead} >= 1024 * 1024; # arbitrary
+            # ->put_writeout clears {read_ahead}, so we run it after we need that
             $self->put_writeout if $self->{read_ahead} >= 8192; # arbitrary
         },
         on_disconnect => sub {
@@ -268,6 +271,8 @@ sub handle_put_chunked {
         },
     }});
 
+    $self->start_put_open($disk_path, $filename);
+
     return 1;
 }
 
@@ -331,6 +336,9 @@ sub event_read_put {
     $self->{content_length_remain} -= $clen;
 
     if ($self->{content_length_remain}) {
+        # Reading too far ahead of our AIO subsystem will cause us to buffer it in memory.
+        $self->watch_read(0) if $self->{read_ahead} >= 1024 * 1024; # arbitrary
+        # ->put_writeout clears {read_ahead}, so we run it after we need that
         $self->put_writeout if $self->{read_ahead} >= 8192; # arbitrary
     } else {
         # now, if we've filled the content of this put, we're done
@@ -395,6 +403,14 @@ sub start_put_open {
         $self->{put_pos}         = 0;
         $self->{put_fh_filename} = "$path/$file";
 
+        # We just opened the file, haven't read_ahead any bytes, are expecting 0 bytes for read and we're
+        # not in chunked mode, so close the file immediately, we're done.
+        unless ($self->{read_ahead} || $self->{content_length_remain} || $self->{chunked_upload_state}) {
+            # FIXME this should be done through AIO
+            $self->put_close;
+            return;
+        }
+
         $self->put_writeout;
     });
 }
@@ -416,6 +432,9 @@ sub put_writeout {
     $self->{read_buf}   = [];
     $self->{read_ahead} = 0;
 
+    # After copying out and clearing the buffer, turn reads back on again to fill up another buffer.
+    $self->watch_read(1) if $self->{content_length_remain} || $self->{chunked_upload_state};
+
     # okay, file is open, write some data
     $self->{put_in_progress} = 1;
 
