@@ -36,6 +36,7 @@ use fields ('service',             # Perlbal::Service object
 
             # service selector parent
             'selector_svc',        # the original service from which we came
+            'is_ssl',              # Is this socket SSL attached (restricted operations)
             );
 
 use Fcntl ':mode';
@@ -76,6 +77,7 @@ sub new {
     $self->{requests}        = 0;
     $self->{scratch}         = {};
     $self->{selector_svc}    = $selector_svc;
+    $self->{is_ssl}          = 0;
 
     $self->state('reading_headers');
 
@@ -83,6 +85,15 @@ sub new {
     return $self;
 }
 
+sub new_from_base {
+    my $class = shift;
+    my Perlbal::ClientHTTPBase $cb = shift;    # base object
+    Perlbal::Util::rebless($cb, $class);
+
+    $cb->handle_request;
+    return $cb;
+}
+
 sub close {
     my Perlbal::ClientHTTPBase $self = shift;
 
@@ -109,6 +120,7 @@ sub setup_keepalive {
     # now get the headers we're using
     my Perlbal::HTTPHeaders $reshd = $_[1];
     my Perlbal::HTTPHeaders $rqhd = $self->{req_headers};
+    my $override_value = $_[2];
 
     # for now, we enforce outgoing HTTP 1.0
     $reshd->set_version("1.0");
@@ -117,13 +129,14 @@ sub setup_keepalive {
     # we respect for persist_client
     my $svc = $self->{selector_svc} || $self->{service};
     my $persist_client = $svc->{persist_client} || 0;
+    $persist_client = $override_value if defined $override_value;
     print "  service's persist_client = $persist_client\n" if Perlbal::DEBUG >= 3;
 
     # do keep alive if they sent content-length or it's a head request
     my $do_keepalive = $persist_client && $rqhd->req_keep_alive($reshd);
     if ($do_keepalive) {
         print "  doing keep-alive to client\n" if Perlbal::DEBUG >= 3;
-        my $timeout = $self->{service}->{persist_client_timeout};
+        my $timeout = $self->{service}->{persist_client_idle_timeout};
         $reshd->header('Connection', 'keep-alive');
         $reshd->header('Keep-Alive', $timeout ? "timeout=$timeout, max=100" : undef);
     } else {
@@ -139,7 +152,17 @@ sub setup_keepalive {
 
 # overridden here from Perlbal::Socket to use the service value
 sub max_idle_time {
-    return $_[0]->{service}->{persist_client_timeout};
+    my Perlbal::ClientHTTPBase $self = shift;
+    if ($self->state eq 'persist_wait') {
+        return $self->{service}->{persist_client_idle_timeout};
+    } else {
+        return $self->{service}->{idle_timeout};
+    }
+}
+
+# Called when this client is entering a persist_wait state, but before we are returned to base.
+sub persist_wait {
+    
 }
 
 # called when we've finished writing everything to a client and we need
@@ -192,6 +215,8 @@ sub http_response_sent {
     $self->{post_sendfile_cb} = undef;
     $self->state('persist_wait');
 
+    $self->persist_wait;
+
     if (my $selector_svc = $self->{selector_svc}) {
         if (! $selector_svc->run_hook('return_to_base', $self)){
             $selector_svc->return_to_base($self);
@@ -215,11 +240,41 @@ sub reproxy_fh {
         $self->{reproxy_fh} = $fh;
         $self->{reproxy_file_offset} = 0;
         $self->{reproxy_file_size} = $size;
-        # call hook that we're reproxying a file
-        return $fh if $self->{service}->run_hook("start_send_file", $self);
-        # turn on writes (the hook might not have wanted us to)
-        $self->watch_write(1);
-        return $fh;
+
+        my $is_ssl_webserver = ( $self->{service}->{listener}->{sslopts} &&
+                               ( $self->{service}->{role} eq 'web_server') );
+
+        unless ($is_ssl_webserver) {
+            # call hook that we're reproxying a file
+            return $fh if $self->{service}->run_hook("start_send_file", $self);
+            # turn on writes (the hook might not have wanted us to)
+            $self->watch_write(1);
+            return $fh;
+        } else { # use aio_read for ssl webserver instead of sendfile
+
+            print "webserver in ssl mode, sendfile disabled!\n"
+                if $Perlbal::DEBUG >= 3;
+
+            # turn off writes
+            $self->watch_write(0);
+            #create filehandle for reading
+            my $data = '';
+            Perlbal::AIO::aio_read($self->reproxy_fh, 0, 2048, $data, sub {
+                # got data? undef is error
+                return $self->_simple_response(500) unless $_[0] > 0;
+
+                # seek into the file now so sendfile starts further in
+                my $ld = length $data;
+                sysseek($self->{reproxy_fh}, $ld, &POSIX::SEEK_SET);
+                $self->{reproxy_file_offset} = $ld;
+                # reenable writes after we get data
+                $self->tcp_cork(1); # by setting reproxy_file_offset above,
+                                    # it won't cork, so we cork it
+                $self->write($data);
+                $self->watch_write(1);
+            });
+            return 1;
+        }
     }
 
     return $self->{reproxy_fh};
@@ -235,8 +290,17 @@ sub event_read {
         if $self->{req_headers};
 
     my $hd = $self->read_request_headers;
+    $self->handle_request;
+}
+
+sub handle_request {
+    my Perlbal::ClientHTTPBase $self = shift;
+    my Perlbal::HTTPHeaders $hd = $self->{req_headers};
+
     return unless $hd;
 
+    $self->check_req_headers;
+
     return if $self->{service}->run_hook('start_http_request', $self);
 
     # we must stop watching for events now, otherwise if there's
@@ -264,6 +328,19 @@ sub event_read {
     }
 }
 
+sub reproxy_file_done {
+    my Perlbal::ClientHTTPBase $self = shift;
+    return if $self->{service}->run_hook('reproxy_fh_finished', $self);
+    # close the sendfile fd
+    CORE::close($self->{reproxy_fh});
+    $self->{reproxy_fh} = undef;
+    if (my $cb = $self->{post_sendfile_cb}) {
+        $cb->();
+    } else {
+        $self->http_response_sent;
+    }
+}
+
 # client is ready for more of its file.  so sendfile some more to it.
 # (called by event_write when we're actually in this mode)
 sub event_write_reproxy_fh {
@@ -271,12 +348,47 @@ sub event_write_reproxy_fh {
 
     my $remain = $self->{reproxy_file_size} - $self->{reproxy_file_offset};
     $self->tcp_cork(1) if $self->{reproxy_file_offset} == 0;
+    $self->watch_write(0);
+
+    if ($self->{is_ssl}) { # SSL (sendfile does not do SSL)
+        return if $self->{closed};
+        if ($remain <= 0) { #done
+            print "REPROXY SSL done\n" if Perlbal::DEBUG >= 2;
+            $self->reproxy_file_done;
+            return;
+        }
+        # queue up next read
+        Perlbal::AIO::set_file_for_channel($self->{reproxy_file});
+        my $len = $remain > 4096 ? 4096 : $remain; # buffer size
+        my $buffer = '';
+        Perlbal::AIO::aio_read(
+            $self->{reproxy_fh},
+            $self->{reproxy_file_offset},
+            $len,
+            $buffer,
+            sub {
+                return if $self->{closed};
+                # we have buffer to send
+                my $rv = $_[0]; # arg is result of sysread
+                if (!defined($rv) || $rv <= 0) { # read error
+                    # sysseek is called after sysread so $! not valid
+                    $self->close('sysread_error');
+                    print STDERR "Error w/ reproxy sysread\n";
+                    return;
+                }
+                $self->{reproxy_file_offset} += $rv;
+                $self->tcp_cork(1); # by setting reproxy_file_offset above,
+                                    # it won't cork, so we cork it
+                $self->write($buffer); # start socket send
+                $self->watch_write(1);
+            } 
+        );
+        return;
+    }
 
     # cap at 128k sendfiles
     my $to_send = $remain > 128 * 1024 ? 128 * 1024 : $remain;
 
-    $self->watch_write(0);
-
     my $postread = sub {
         return if $self->{closed};
 
@@ -296,17 +408,7 @@ sub event_write_reproxy_fh {
         $self->{reproxy_file_offset} += $sent;
 
         if ($sent >= $remain) {
-            return if $self->{service}->run_hook('reproxy_fh_finished', $self);
-
-            # close the sendfile fd
-            CORE::close($self->{reproxy_fh});
-
-            $self->{reproxy_fh} = undef;
-            if (my $cb = $self->{post_sendfile_cb}) {
-                $cb->();
-            } else {
-                $self->http_response_sent;
-            }
+            $self->reproxy_file_done;
         } else {
             $self->watch_write(1);
         }
@@ -776,6 +878,45 @@ sub send_response {
     return $self->_simple_response(@_);
 }
 
+sub send_full_response {
+    my Perlbal::ClientHTTPBase $self = shift;
+    my $code = shift;
+    my $headers = shift || [];
+    my $bref = ref($_[0]) eq 'SCALAR' ? shift : \shift;
+    my $options = shift || {};
+
+    my $res = $self->{res_headers} = Perlbal::HTTPHeaders->new_response($code);
+
+    while (@$headers) {
+        my ($name, $value) = splice @$headers, 0, 2;
+        $res->header($name, $value);
+    }
+
+    if ($code == 204 || $code == 304) {
+        $res->header('Content-Length', undef);
+        $bref = \undef;
+    } elsif (defined $$bref) {
+        $res->header('Content-Length', length($$bref));
+    }
+
+    $res->header('Server', 'Perlbal'); # Tunable?
+    # $res->header('Date', # We should do this
+
+    $self->setup_keepalive($res, $options->{persist_client});
+
+    $self->state('xfer_resp');
+    $self->tcp_cork(1);  # cork writes to self
+    $self->write($res->to_string_ref);
+
+    if (defined $$bref && $self->{req_headers} && $self->{req_headers}->request_method ne 'HEAD') {
+        # don't write body for head requests
+        $self->write($bref);
+    }
+
+    $self->write(sub { $self->http_response_sent; });
+    return 1;
+}
+
 # method that sends a 500 to the user but logs it and any extra information
 # we have about the error in question
 sub system_error {
@@ -792,12 +933,20 @@ sub system_error {
 sub event_err {  my $self = shift; $self->close('error'); }
 sub event_hup {  my $self = shift; $self->close('hup'); }
 
+sub _sock_port {
+    my $name = $_[0];
+    my $port = eval { (Socket::sockaddr_in($name))[0] };
+    return $port unless $@;
+    # fallback to IPv6:
+    return (Socket6::unpack_sockaddr_in($name))[0];
+}
+
 sub as_string {
     my Perlbal::ClientHTTPBase $self = shift;
 
     my $ret = $self->SUPER::as_string;
     my $name = $self->{sock} ? getsockname($self->{sock}) : undef;
-    my $lport = $name ? (Socket::sockaddr_in($name))[0] : undef;
+    my $lport = $name ? _sock_port($name) : undef;
     my $observed = $self->observed_ip_string;
     $ret .= ": localport=$lport" if $lport;
     $ret .= "; observed_ip=$observed" if defined $observed;
